from .abc import BaseComponent
from .partials import PartialEmoji
from .exceptions import TooManyComponents, InvalidMessageButtonStyle, InvalidArgumentType, TooManySelectMenuOptions, LabelIsTooBig
from typing import (
    Union,
    List,
    Optional
)

class MessageSelectMenuOption:
    def __init__(self, label: str, value: str, description: Optional[str], emoji: Optional[PartialEmoji], default: Optional[bool]):
        self.settings = {
            "label": label,
            "value": value,
            "description": description or None,
            "emoji": emoji or None,
            "default": default or None            
        }
    
    def __repr__(self):
        return self.settings

class MessageSelectMenu(BaseComponent):
    def __init__(self):
        self.settings = {
            "options": [], 
            "type": 3,
            "min_values": 1,
            "max_values": 1,
            "disabled": False
        }
    
    def __repr__(self):
        return self.settings

    def add_options(self, options: List[MessageSelectMenuOption]):
        for option in options:
            
            if len(self.settings["options"] > 25):
                raise TooManySelectMenuOptions("You can only have 25 options in a select menu.")
            
            self.settings["options"].append(option.data)
        
    def set_placeholder(self, placeholder: str):
        if not isinstance(placeholder, str):
            raise InvalidArgumentType("Placeholder must be a string.")
        
        self.settings["placeholder"] = placeholder
    
    def set_min_values(self, min: int):
        if not isinstance(min, int):
            raise InvalidArgumentType("Min must be an integer.")
        
        self.options["min_values"] = min
        
    def set_max_values(self, max: int):
        if not isinstance(max, int):
            raise InvalidArgumentType("Max must be an integer.")
        
        self.options["max_values"] = max    


class MessageButton(BaseComponent):
    def __init__(self):
        self.settings = {
            "type": 2,
            "style": 1,
            "label": "Click me!",
            "emoji": None,
            "disabled": False
        }
    
    def __repr__(self):
        return self.settings

    def set_label(self, label: str):

        if not isinstance(label, str):
            raise InvalidArgumentType("Label must be a string.")
        
        if len(label) > 80:
            raise LabelIsTooBig("Label must be 80 characters or less.")
            
        self.settings["label"] = label
    
    def set_style(self, style: Union[int, str]):
        valid_styles = {
            "PRIMARY": 1,
            "SECONDARY": 2,
            "SUCCESS": 3,
            "DANGER": 4,
            "LINK": 5
        }
        if isinstance(style, str):
            if style.upper() not in valid_styles:
                raise InvalidMessageButtonStyle("Invalid button style. Style must be one of PRIMARY, SECONDARY, LINK, DANGER, or SUCCESS.")
            self.settings["style"] = style.upper()
            
        elif isinstance(style, int):
            if style not in valid_styles.values():
                raise InvalidMessageButtonStyle("Invalid button style. Style must be in range 1 to 5 inclusive.")
    
    def set_emoji(self, emoji: Union[PartialEmoji, dict]):

        if isinstance(emoji, dict):
            self.settings["emoji"] = emoji
            return

        elif isinstance(emoji, PartialEmoji):
            self.settings["emoji"] = emoji.data
            return
        raise InvalidArgumentType("Emoji must be a PartialEmoji or a dict that represents a PartialEmoji.")
                    
    def set_url(self, url: str):
        
        if not isinstance(url, str):
            raise InvalidArgumentType("Url must be a string.")
        
        self.settings["url"] = url
        self.settings["style"] = 5
        
        

class MessageActionRow:
    def __init__(self, components: Optional[List[Union[MessageButton, MessageSelectMenu]]]):
        self.settings = {
            "type": 1,
            "components": components
        }

    def __repr__(self):
        return self.settings
        
    def add_components(self, components: List[Union[MessageButton, MessageSelectMenu]]):
        buttons = 0
        for component in self.settings["components"]:
            if type(component) == MessageButton:
                buttons += 1
            
            elif buttons >= 5:
                raise TooManyComponents("You can only have 5 buttons per row.")
            
            elif type(component) == MessageSelectMenu:
                raise TooManyComponents("You can only have 1 select menu per row. No buttons along that select menu.")
        self.settings["components"].append(components)


class DiscordAPIError(Exception): # 
    ...
    
class InvalidToken(Exception):
    ...

class BadRequest400(Exception):
    ...

class Unauthorized401(Exception):
    ...

class Forbidden403(Exception):
    ...

class NotFound404(Exception):
    ...

class MethodNotAllowed405(Exception):
    ...
    
class Ratelimited429(Exception):
    ...

class GateawayUnavailable502(Exception):
    ...
    
class InternalServerError5xx(Exception):
    ...
    
class TooManyComponents(Exception):
    ...

class InvalidMessageButtonStyle(Exception):
    ...

class CustomIdIsTooBig(Exception):
    ...

class InvalidArgumentType(Exception):
    ...

class TooManySelectMenuOptions(Exception):
    ...
    
class LabelIsTooBig(Exception):
    ...
    
class ThreadArchived(Exception):
    ...



from typing import (
    Optional,
    List,
    Union
)
from .abc import BaseSlashCommandOption
from .application import ApplicationCommand

class Subcommand(BaseSlashCommandOption):
    def __init__(self, *, name: str, description: Optional[str] = None, required: bool = False):
        super().__init__(name=name, description=description, required=required)
        self.settings["type"] = 1
        
class SubCommandGroup(BaseSlashCommandOption):
    def __init__(self, *, name: str, description: Optional[str] = None, required: bool = False):
        super().__init__(name=name, description=description, required=required)
        self.settings["type"] = 2
        
class StringOption(BaseSlashCommandOption):
    def __init__(self, *, name: str, description: Optional[str] = None, required: bool = False):
        super().__init__(name=name, description=description, required=required)
        self.settings["type"] = 3
        
class IntegerOption(BaseSlashCommandOption):
    def __init__(self, *, name: str, description: Optional[str] = None, required: bool = False):
        super().__init__(name=name, description=description, required=required)
        self.settings["type"] = 4
        
class BooleanOption(BaseSlashCommandOption):
    def __init__(self, *, name: str, description: Optional[str] = None, required: bool = False):
        super().__init__(name=name, description=description, required=required)
        self.settings["type"] = 5
        
class UserOption(BaseSlashCommandOption):
    def __init__(self, *, name: str, description: Optional[str] = None, required: bool = False):
        super().__init__(name=name, description=description, required=required)
        self.settings["type"] = 6
        
class ChannelOption(BaseSlashCommandOption):
    def __init__(self, *, name: str, description: Optional[str] = None, required: bool = False):
        super().__init__(name=name, description=description, required=required)
        self.settings["type"] = 7
        
class RoleOption(BaseSlashCommandOption):
    def __init__(self, *, name: str, description: Optional[str] = None, required: bool = False):
        super().__init__(name=name, description=description, required=required)
        self.settings["type"] = 8

class MentionableOption(BaseSlashCommandOption):
    def __init__(self, *, name: str, description: Optional[str] = None, required: bool = False):
        super().__init__(name=name, description=description, required=required)
        self.settings["type"] = 9
    
class NumberOption(BaseSlashCommandOption):
    def __init__(self, *, name: str, description: Optional[str] = None, required: bool = False):
        super().__init__(name=name, description=description, required=required)
        self.settings["type"] = 10

class SlashCommandOptionChoice:
    def __init__(self, * name: str, value: Union[float, int, str]):
        self.settings = {
            "name": name,
            "value": value
        }


class SlashCommand(ApplicationCommand):
    def __init__(self, data: dict):
        super().__init__(data)
        self.options: Optional[List[Union[Subcommand, SubCommandGroup, StringOption, IntegerOption, BooleanOption, UserOption, ChannelOption, RoleOption, MentionableOption, NumberOption]]] = data["options"] or None # Return the type hinted class later this will take too long and is very tedious, I'll probably get Copilot to do it for me lmao


from .abc import BaseInteraction
from websocket import WebSocket
from asyncio import run
from .slash_command import Subcommand, SubCommandGroup, StringOption, IntegerOption, BooleanOption, UserOption, ChannelOption, RoleOption, MentionableOption, NumberOption
from typing import (
    List,
    Union
)
from time import sleep
from json import loads, dumps
from threading import _start_new_thread

class WebsocketClient:
    def __init__(self, token: str, intents: int):

        self.EVENT = 0
        self.HEARTBEAT = 1
        self.IDENTIFY = 2
        self.PRESENCE_UPDATE = 3
        self.VOICE_STATE_UPDATE = 4
        self.RESUME = 6
        self.RECONNECT = 7
        self.REQUEST_GUILD_MEMBERS = 8
        self.INVALID_SESSION = 9
        self.HELLO = 10
        self.HEARTBEAT_ACK = 11

        self.ws = WebSocket()
        self.token = token
        self.intents = intents
        self.interval = None
        self.session_id = None
        self.events = {
            "interaction": self.interaction
        }
        self.events_to_handle = {
            
        }
        self.commands = {}
        self.hearbeats = []
        self.average_latency = 0
        
    async def interaction(self, interaction: BaseInteraction):
        await self.commands[interaction.command_name]["callback"](interaction)
        
    
    def heartbeat(self):
        while True:
            sleep(self.interval / 1000)
            self.ws.send(dumps({"op": self.HEARTBEAT, "d": "null"}))
            event = self.receive_event()
            if event:
                if event["op"] == self.HEARTBEAT_ACK:
                    self.heartbeats.append(event["d"])
                    self.sequence = event["s"]
            print("Sent heartbeat!")
    
    def receive_event(self):
        response = self.ws.recv()
        if response:
            return loads(response)
    
    def handle_event(self, event_name: str, data: dict):
        try:
            function = self.events[event_name]
            run(function(data))
        except KeyError:
            print(f"You have not registered an event handler for {event_name} but still receive the event. Either make a handler or remove the intent to view this event if possible.") # Someone change this to logger

    def infinitely_retreive_events(self):
        while True:
            event = self.receive_event()
            if event:
                if event["op"] == self.EVENT:
                    self.handle_event(event["t"].lower(), event["d"])
                elif event["op"] == self.HEARTBEAT:
                    self.ws.send(dumps({"op": self.HEARTBEAT, "d": self.sequence or "null"}))

    def event(self, func):
        def register_event():
            self.events[func.__name__.lower().replace("on_")] = func
        return register_event
    
    def command(self, *, name: str, description: str, guild_ids: List[str], options:Union[Subcommand, SubCommandGroup, StringOption, IntegerOption, BooleanOption, UserOption, ChannelOption, RoleOption, MentionableOption, NumberOption]):
        def register_slash_command(func):
            self.commands[func.__name__] = {"callback": func, "name": name, "description": description, "guild_ids": guild_ids, "options": options}
        return register_slash_command

    def login(self):
        self.ws.connect("wss://gateway.discord.gg/?v=9&encoding=json")
        event = self.receive_event()
        print(event)
        self.interval = event["d"]["heartbeat_interval"]
        self.ws.send(
            dumps({
            "op": self.IDENTIFY,
            "d": {
                "token": self.token,
                "intents": self.intents,
                "properties": {
                    "$os": "linux",
                    "$browser": "EpikCord",
                    "$device": "EpikCord"
                    },
                }
            })
        )
        _start_new_thread(self.heartbeat, ())
        
        self.infinitely_retreive_events()


class Overwrite:
    def __init__(self, data: dict):
        self.id: str = data["id"]
        self.type: int = data["type"]
        self.allow: str = data["allow"]
        self.deny: str = data["deny"]


from typing import (
    Optional
)
from .channels import GuildChannel, Thread
from .emoji import Emoji
from .user import User
from .member import Member
from .client import Client
from .role import Role
from typing import (
    List
)

class Guild:
    def __init__(self, client: Client, data: dict):
        self.client = client
        self.data: dict = data
        self.id: str = data["id"]
        self.name: str = data["name"]
        self.icon: Optional[str] = data["icon"] or None
        self.icon_hash: Optional[str] = data["icon_hash"] or None
        self.splash: Optional[str] = data["splash"] or None
        self.discovery_splash: Optional[str] = data["discovery_splash"] or None
        self.owner_id: str = data["owner_id"]
        self.permissions: str = data["permissions"]
        self.afk_channel_id: str = data["afk_channel_id"]
        self.afk_timeout: int = data["afk_timeout"]
        self.verification_level: str = "NONE" if data["verification_level"] == 0 else "LOW" if data["verification_level"] == 1 else "MEDIUM" if data["verification_level"] == 2 else "HIGH" if data["verification_level"] == 3 else "VERY_HIGH"
        self.default_message_notifications: str = "ALL" if data["default_message_notifications"] == 0 else "MENTIONS" 
        self.explicit_content_filter: str = "DISABLED" if data["explicit_content_filter"] == 0 else "MEMBERS_WITHOUT_ROLES" if data["explicit_content_filter"] == 1 else "ALL_MEMBERS"
        self.roles: List[Role] = [Role(role) for role in data["roles"]]
        self.emojis: List[Emoji] = [Emoji(emoji) for emoji in data["emojis"]]
        self.features: List[str] = data["features"]
        self.mfa_level: str = "NONE" if data["mfa_level"] == 0 else "ELEVATED"
        self.application_id: Optional[str] = data["application_id"] or None
        self.system_channel_id: Optional[str] = data["system_channel_id"] or None
        self.system_channel_flags: int = data["system_channel_flags"]
        self.rules_channel_id: Optional[int] = data["rules_channel_id"] or None
        self.joined_at: Optional[str] = data["joined_at"] or None
        self.large: bool = data["large"]
        self.unavailable: bool = data["unavailable"]
        self.member_count: int = data["member_count"]
        # self.voice_states: List[dict] = data["voice_states"]
        self.members: List[Member] = [Member(member) for member in data["members"]]
        self.channels: List[GuildChannel] = [GuildChannel(channel) for channel in data["channels"]]
        self.threads: List[Thread] = [Thread(thread) for thread in data["threads"]]
        self.presences: List[dict] = data["presences"]
        self.max_presences: int = data["max_presences"]
        self.max_members: int = data["max_members"]
        self.features
        
class GuildScheduledEvent:
    def __init__(self, client: Client, data: dict):
        self.id: str = data["id"]
        self.client = client
        self.guild_id: str = data["guild_id"]
        self.channel_id: Optional[str] = data["channel_id"] or None
        self.creator_id: Optional[str] = data["creator_id"] or None
        self.name: str = data["name"]
        self.description: Optional[str] = data["description"] or None
        self.scheduled_start_time: str = data["scheduled_start_time"]
        self.scheduled_end_time: Optional[str] = data["scheduled_end_time"] or None
        self.privacy_level: int = data["privacy_level"]
        self.status: str = "SCHEDULED" if data["status"] == 1 else "ACTIVE" if data["status"] == 2 else "COMPLETED" if data["status"] == 3 else "CANCELLED"
        self.entity_type: str = "STAGE_INSTANCE" if data["status"] == 1 else "VOICE" if data["status"] == 2 else "EXTERNAL"
        self.entity_id: str = data["entity_id"]
        self.entity_metadata: dict = data["entity_metadata"]
        self.creator: Optional[User] = User(data["creator"]) or None
        self.user_count: Optional[int] = data["user_count"] or None


from EpikCord import *


client = Client("ODk4MjY4MjI3MjE2NTA2OTQw.YWhveg.PeQ0IZhJKEwNeoh8wWJ4PcK5TVQ", 513)

@client.event
async def interaction(interaction: dict):
    interaction = BaseInteraction(client, interaction)
    await interaction.reply({"content": "Hello World"})
    
client.login()


from typing import (
    List
)
from .section import Section
from .channels import TextBasedChannel
from .abc import BaseChannel
from .ws import WebsocketClient
from asyncio import run
from .member import Member
from .user import User
from .application import Application
from .route import Route
from .abc import Messageable
from aiohttp import ClientSession


class Client(WebsocketClient):

    def __init__(self, token: str, intents: int = 0, **options):
        super().__init__(token, intents)
        
        self.channels: List[Messageable] = []
        
        self.options: dict = options

        self.http = ClientSession(headers = {"Authorization": f"Bot {token}"}, base_url="https://discord.com/api/v9/")
        self.api = Route
        self.application: Application = Application(self, self.user) # Processes whatever it can        

    async def get_channel(self, channel_id: str):
        try:
            if self.channels[channel_id]:
                return self.channels[channel_id]
        except KeyError:
            response = await self.http.get(f"channels/{channel_id}")
            data = await response.json()
            return TextBasedChannel(data)
    
    def add_section(self, section: Section):
        for name, command_object in section.commands:
            self.client.commands[name] = command_object

        for event_name, event_func in section.events:
            self.events[event_name.lower().replace("on_")] = event_func
        
        # Successfully extracted all the valuable stuff from the section        

# class ClientUser(User):
    
#     def __init__(self, client: Client, data: dict):
#         super().__init__(data)
    
#     async def fetch(self):
#         response = await self.client.http.get("users/@me")
#         data = await response.json()
#         super().__init__(data) # Reinitialse the class with the new data.
    
# class ClientGuildMember(Member):
#     def __init__(self, client: Client,data: dict):
#         super().__init__(data)


from .team import Team
from .partials import PartialUser
from aiohttp import ClientResponse
from typing import (
    Optional
)
from .client import Client

class Application:
    def __init__(self, client: Client, data: dict):
        self.id: str = data["id"]
        self.client: Client = client
        self.name: str = data["name"]
        self.icon: Optional[str] = data["icon"] or None
        self.description: str = data["description"]
        self.rpc_origins: Optional[list] = data["rpc_origins"] or None
        self.bot_public: bool = data["bot_public"]
        self.bot_require_code_grant: bool = data["bot_require_code_grant"]
        self.terms_of_service_url: Optional[str] = data["terms_of_service"] or None
        self.privacy_policy_url: Optional[str] = data["privacy_policy"] or None
        self.owner: PartialUser = PartialUser(data["user"])
        self.summary: str = data["summary"]
        self.verify_key: str = data["verify_key"]
        self.team: Optional[Team] = Team(data["team"]) or None
        self.cover_image: Optional[str] = data["cover_image"] or None
        self.flags: int = data["flags"]

    async def fetch(self):
        response: ClientResponse = await self.client.api(self, "oauth2/applications/@me").request("GET")
        data: dict = await response.json()
        self.application = Application(data)
        
class ApplicationCommand:
    def __init__(self, data: dict):
        self.id: str = data["id"]
        self.type: int = data["type"]
        self.application_id: str = data["application_id"]
        self.guild_id: Optional[str] = data["guild_id"] or None
        self.name: str = data["name"]
        self.description: str = data["description"]
        self.default_permissions: bool = data["default_permissions"]
        self.version: str = data["version"]


from typing import (
    List,
    Optional
)
from .client import Client
from .user import User
from .message import Message
from .member import GuildMember
from .exceptions import (
    InvalidArgumentType,
    CustomIdIsTooBig
)

class Messageable:
    def __init__(self, client: Client, channel_id: str):
        self.channel_id: str = channel_id
        self.client = client
        
    async def fetch_messages(self,*, around: Optional[str] = None, before: Optional[str] = None, after: Optional[str] = None, limit: Optional[int] = None) -> List[Message]:
        response = await self.client.http.get(f"channels/{self.id}/messages", params={"around": around, "before": before, "after": after, "limit": limit})
        data = await response.json()
        return [Message(message) for message in data]
    
    async def fetch_message(self,*, message_id: str) -> Message:
        response = await self.client.http.get(f"channels/{self.id}/messages/{message_id}")
        data = await response.json()
        return Message(data)

    async def send(self, message_data: dict) -> Message:
        response = await self.client.http.post(f"channels/{self.id}/messages", data=message_data)
        return Message(await response.json())
    
class BaseSlashCommandOption:
    def __init__(self, *, name: str, description: str, required: bool = False):
        self.settings = {
            "name": name,
            "description": description,
            "required": required
        }
        # People shouldn't use this class, this is just a base class for other options
    
class BaseChannel:
    def __init__(self, client, data: dict):
        self.id: str = data["id"]
        self.client: Client = client
        self.type = data["type"]

class BaseComponent:
    def __init__(self):
        self.settings = {}

    def set_custom_id(self, custom_id: str):
        
        if not isinstance(custom_id, str):
            raise InvalidArgumentType("Custom Id must be a string.")
        
        elif len(custom_id) > 100:
            raise CustomIdIsTooBig("Custom Id must be 100 characters or less.")

        self.settings["custom_id"] = custom_id
        
class BaseInteraction:
    def __init__(self, client, data: dict):
        self.id: str = data["id"]
        self.client = client
        self.application_id: int = data["application_id"]
        self.type: int = data["type"]
        self.data: Optional[dict] = data["data"] or None
        self.guild_id: Optional[str] = data["guild_id"] or None
        self.channel_id: Optional[str] = data["channel_id"] or None
        self.member: Optional[GuildMember] = GuildMember(data["member"]) or None
        self.user: Optional[User] = User(data["user"]) or None
        self.token: str = data["token"]
        self.version: int = data["version"]
        self.message: Optional[Message] = Message(data["message"]) or None
        self.locale: Optional[str] = data["locale"] or None
        self.guild_locale: Optional[str] = data["guild_locale"] or None
        
    def is_ping(self):
        return self.type == 1
    
    def is_application_command(self):
        return self.type == 2
    
    def is_message_component(self):
        return self.type == 3
    
    def is_autocomplete(self):
        return self.type == 4
    
    async def reply(self, message_data: dict):
        response = await self.client.http.post(f"/interactions/{self.id}/{self.token}/callback", data=message_data)
        return await response.json()
    
    async def fetch_reply(self):
        response = await self.client.http.get(f"/webhooks/{self.application_id}/{self.token}/{self.token}/messages/@original")
        return await response.json()
    
    async def edit_reply(self, message_data: dict):
        response = await self.client.http.patch(f"/webhooks/{self.application_id}/{self.token}/messages/@original", data=message_data)
        return await response.json()
    
    async def delete_reply(self):
        response = await self.client.http.delete(f"/webhooks/{self.application_id}/{self.token}/messages/@original")
        return await response.json()
    
    async def followup(self, message_data: dict):
        response = await self.client.http.post(f"/webhooks/{self.application_id}/{self.token}", data=message_data)
        return await response.json()
    
    async def fetch_followup_message(self, message_id: Message.id):
        response = await self.client.http.get(f"/webhooks/{self.application_id}/{self.token}/messages/{message_id}")
        return await response.json()
    
    async def edit_followup(self, message_id: Message.id, message_data):
        response = await self.client.http.patch(f"/webhooks/{self.application_id}/{self.token}/messages/{message_id}", data=message_data)
        return await response.json()
    
    async def delete_followup(self, message_id: Message.id):
        response = await self.client.http.delete(f"/webhooks/{self.application_id}/{self.token}/messages/{message_id}")
        return await response.json()
    
    